package pa;

import java.nio.IntBuffer;
import java.util.Random;

import org.lwjgl.BufferUtils;
import org.lwjgl.PointerBuffer;
import org.lwjgl.opencl.CL10;
import org.lwjgl.opencl.CLCommandQueue;
import org.lwjgl.opencl.CLContext;
import org.lwjgl.opencl.CLKernel;
import org.lwjgl.opencl.CLMem;
import org.lwjgl.opencl.CLProgram;

import pa.cl.CLUtil;
import pa.cl.CLUtil.PlatformDevicePair;
import pa.cl.OpenCL;
import pa.util.BufferHelper;
import pa.util.IOUtil;
import pa.util.SizeOf;

public class ScanCompactRT {
    
    private static CLContext context;
    private static CLKernel reduceForward;
    private static CLKernel reduceBackward;
    private static CLKernel compact;
    private static CLCommandQueue queue;
    private static CLProgram program;

    private static CLMem clRayBool;
    private static CLMem clRayIndex;
    private static CLMem clPrefixSum;
    private static CLMem clCompactResult;
    
    private static int width = 4;
    private static int height = 4;
    
    private static IntBuffer rays = BufferUtils.createIntBuffer(width * height);
    private static IntBuffer tmp = BufferUtils.createIntBuffer(width * height);
    
    private static Random r = new Random(0);
    
    private static PointerBuffer gws = new PointerBuffer(1);
    
    private static interface Operation
    {
        void operate(IntBuffer data, int index);
    }
    
    private static class ComputeRays implements Operation
    {
        @Override
        public void operate(IntBuffer data, int index) {
            data.put(index, data.get(index) * r.nextInt(2));
        }
    }
    
    private static class InitRays implements Operation
    {
        @Override
        public void operate(IntBuffer data, int index) {
            data.put(index, 1);
        }
    }
    
    private static class InitIndex implements Operation
    {
        @Override
        public void operate(IntBuffer data, int index) {
            data.put(index, index);
        }
    }
    
    private static void doRays(IntBuffer rays, int max, Operation op)
    {
        for(int i = 0; i < max; ++i)
        {
            op.operate(rays, i);
        }
    }
    
    private static void computeNewRays(IntBuffer rays)
    {
        doRays(rays, rays.capacity(), new ComputeRays());
    }
    
    private static void initRays(IntBuffer rays)
    {
        doRays(rays, rays.capacity(), new InitRays());
    }
    
    private static void download(CLMem clRays, IntBuffer rays, int offset)
    {
        rays.position(0);
        OpenCL.clEnqueueReadBuffer(queue, clRays, 1, offset, rays, null, null);
    }
    
    private static void upload(CLMem clRays, IntBuffer rays, int offset)
    {
        rays.position(0);
        OpenCL.clEnqueueWriteBuffer(queue, clRays, 1, offset, rays, null, null);
    }
    
    private static void fillZero(IntBuffer rays, int offset)
    {
        for(int i = offset; i < rays.capacity(); ++i)
        {
            rays.put(i, 0);
        }
    }
    
    private static int arrangeRays(IntBuffer rayBool, int raysCount)
    {
        System.out.println("-------------");
        gws.put(0, (width * height) / 2);
        System.out.println("RaysCount=" + raysCount);
        
        fillZero(rayBool, raysCount);

        upload(clRayBool, rayBool, 0);
        
        BufferHelper.printBuffer(rayBool, rayBool.capacity(), "Rays");
        
        CL10.clEnqueueCopyBuffer(queue, clRayBool, clPrefixSum, 0, 0, width * height * SizeOf.INTEGER, null, null);

        OpenCL.clEnqueueNDRangeKernel(queue, reduceForward, 1, null, gws, gws, null, null);
        
        download(clPrefixSum, tmp, 0);
        
        int raysCreated = tmp.get(0);
        
        tmp.put(0, 0);
        
        upload(clPrefixSum, tmp, 0);
        
        OpenCL.clEnqueueNDRangeKernel(queue, reduceBackward, 1, null, gws, gws, null, null);
        
        download(clPrefixSum, tmp, 0);
        BufferHelper.printBuffer(tmp, rayBool.capacity(), "Scan");
        
        gws.put(0, raysCount);

        OpenCL.clEnqueueNDRangeKernel(queue, compact, 1, null, gws, gws, null, null);
        
        download(clCompactResult, tmp, 0);
        
        BufferHelper.printBuffer(tmp, rayBool.capacity(), "Compact");
        
        return raysCreated;
    }
    
    public static void main(String[] args)
    {
        CLUtil.createCL();
        
        PlatformDevicePair pair = CLUtil.choosePlatformAndDevice();

        context = OpenCL.clCreateContext(pair.platform, pair.device, null, null);
        queue = OpenCL.clCreateCommandQueue(context, pair.device, OpenCL.CL_QUEUE_PROFILING_ENABLE);
        program = OpenCL.clCreateProgramWithSource(context, IOUtil.readFileContent("kernel/scanCompact.cl"));
        OpenCL.clBuildProgram(program, pair.device, "", null);
        
        reduceForward = OpenCL.clCreateKernel(program, "reduce");
        reduceBackward = OpenCL.clCreateKernel(program, "reduceInverse");
        compact = OpenCL.clCreateKernel(program, "compact");
        
        clRayBool = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE, rays.capacity() * SizeOf.INTEGER);
        clCompactResult = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE, rays.capacity() * SizeOf.INTEGER);
        clPrefixSum = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE, rays.capacity() * SizeOf.INTEGER);
        clRayIndex = OpenCL.clCreateBuffer(context, OpenCL.CL_MEM_READ_WRITE, rays.capacity() * SizeOf.INTEGER);
        
        doRays(tmp, tmp.capacity(), new InitIndex());

        OpenCL.clEnqueueWriteBuffer(queue, clRayIndex, 1, 0, tmp, null, null);

        OpenCL.clSetKernelArg(reduceForward, 0, clPrefixSum);
        OpenCL.clSetKernelArg(reduceBackward, 0, clPrefixSum);
        
        OpenCL.clSetKernelArg(compact, 0, clPrefixSum);
        OpenCL.clSetKernelArg(compact, 1, clCompactResult);
        OpenCL.clSetKernelArg(compact, 2, clRayBool);
        OpenCL.clSetKernelArg(compact, 3, clRayIndex);
        
        int raysCount = width * height;
        
        initRays(rays);
        while(raysCount > 0)
        {
            //calculate hits and colors here, we ignore it but feel free to implement it ;)
            computeNewRays(rays);
            raysCount = arrangeRays(rays, raysCount);
        }
        
        OpenCL.clReleaseMemObject(clRayBool);
        OpenCL.clReleaseMemObject(clRayIndex);
        OpenCL.clReleaseMemObject(clCompactResult);
        OpenCL.clReleaseMemObject(clPrefixSum);
        OpenCL.clReleaseKernel(reduceForward);
        OpenCL.clReleaseKernel(reduceBackward);
        OpenCL.clReleaseKernel(compact);
        OpenCL.clReleaseCommandQueue(queue);
        OpenCL.clReleaseProgram(program);
        OpenCL.clReleaseContext(context);
        
        CLUtil.destroyCL();
    }
}
